/*
       SASM - Simple Assembler For HHS-E Processor
              see http://ds.opdenbrouw.nl/digse2.html

       (c)2020, J. op den Brouw <J.E.J.opdenBrouw@hhs.nl
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>

#ifdef _MSC_VER
#pragma warning(disable : 4996)
#endif

#define VERSION "sasm 0.2"

/* Maximum length of a line */
#define LINELEN 100

/* Declarations for labels */
#define MAXLABELS 128
#define LABELLEN 100
typedef struct {
	char name[LABELLEN];
	int  seen;
	int  value;
} label_t;

label_t label[MAXLABELS];
int labelindex=0;

/* Declaration of instruction split */
typedef struct {
	char labelname[LABELLEN];
	char mnemonic[LABELLEN];
	char op1[LABELLEN];
	char op2[LABELLEN];
	char op3[LABELLEN];
	char comment[LINELEN];
} instruction_t;

instruction_t instruction;

/* Declaration of machine code bit pattern */
typedef struct {
	char sela[LABELLEN];
	char selb[LABELLEN];
	char seld[LABELLEN];
	char acode[LABELLEN];
	char constant[LABELLEN];
	char end;
	char jcode[LABELLEN];
	char address[LABELLEN];
} machinecode_t;

machinecode_t machinecode;

int errors;

/* Reads a line up to len characters */
int getaline(FILE *fp, char *str, int len) {
	int charsread = 0;
	int c;
	int first=0;

	for (c=0; c<len; c++) str[c] = '\0';

	do {
		c=fgetc(fp);
		//printf("char: %c, charsread: %d\n",c, charsread);
		if (c==EOF) {
			if (charsread == 0) {
				charsread = EOF;
			}
			break;
		}
		if (c=='\n') {
			str[charsread] = '\0';
			break;
		}
		if (c=='\r') {
			continue;
		}
		if (charsread<len) {
			str[charsread] = c;
			charsread++;
		} else {
			if (first==0) {
				first++;
				printf("Warning: superfluous characters ignored\n");
			}
		}
	} while (1);
	return charsread;
}

/* Removes leading and trailing white spaces and gobbles multiple white spaces */
void trimaline(char *str) {

	int again,i,j;

	/* tabs to spaces */
	for (j=0;j<strlen(str);j++) {
		if (str[j]=='\t') {
			str[j] = ' ';
		}
	}

	/* convert multiple spaces to one space */
	again = 1;
	while (again==1) {
		again = 0;
		for (j=0; j<strlen(str);j++) {
			if ((str[j] == ' ') && (str[j+1] == ' ')) {
				again = 1;
				for (i=j+1; i < strlen(str); i++) {
					str[i-1] = str[i];
				}
				str[strlen(str)-1] = '\0';
			}
		}
	}
	/* remove trailing space */
	if (str[strlen(str)-1] == ' ') {
		str[strlen(str)-1] = '\0';
	}
	/* Remove leading space */
	if (str[0] == ' ') {
		for (i=1; i < strlen(str); i++) {
			str[i-1] = str[i];
		}
		str[strlen(str)-1] = '\0';
	}
	return;
}

/* Breaks a line into label, mnemonic, ops 1,2,3 and commment */
int parsealine(char *str, instruction_t *instr) {
	int retval = -1;
	int found = -1;
	int i;
	int j;
	int k;

	/* Clear instruction struct */
	memset(&instruction, 0, sizeof instruction);
	/* printf("Sizeof instruction: %d\n", sizeof instruction); */

	/* Search for : as found in a label*/
	for (i=0; str[i]!='\0'; i++) {
		if (str[i]==':') { found=i; break; }
	}

	/* Check for presence of label */
	if (found>LABELLEN) {
		printf("ERROR: label name too long, skipping!\n");
	} else if (found == 0) {
		printf("ERROR: Empty label, skipping!\n");
	} else if (found>0) {
		strncpy(instr->labelname, str, found);
		instr->labelname[found] = '\0';
	}

	/* If label, then jump over it */
	if (found>=0) { i=found+1; } else { i=0; };
	/*printf("i: %d\n", i); */

	/* Label only line */
	if (str[i] == '\0') {
		//printf("INFO: only label\n");
		return 0;
	}

	/* Hunt for mnemonic */
	if (str[i] == ' ') { i++; }
	for (j=i; j<strlen(str); j++) {
		if (str[j] == ' ' || str[j] == ';' || str[j] == '\0') { break; }
	}
	if (j>i) {
		strncpy(instr->mnemonic, &str[i], j-i);
		//printf("Copied mnemonic...\n");
		i=j;
	}
    /* Make mnemonic lowercase */
	for (k=0; k<strlen(instr->mnemonic); k++) {
		instr->mnemonic[k] = tolower(instr->mnemonic[k]);
	}

	/* Hunt for operand 1 */
	if (str[i] == ' ') { i++; }
	for (j=i; j<strlen(str); j++) {
		if (str[j] == ' ' || str[j] == ';' || str[j] == ',' || str[j] == '\0') { break; }
	}
	if (j>i) {
		strncpy(instr->op1, &str[i], j-i);
		//printf("Copied operand1...\n");
		i=j;
	}
	//printf("i, j: %d, %d\n", i, j);

	/* Hunt for operand 2 */
	if (str[i] == ',') { i++; }
	for (j=i; j<strlen(str); j++) {
		if (str[j] == ' ' || str[j] == ';' || str[j] == ',' || str[j] == '\0') { break; }
	}
	if (j>i) {
		strncpy(instr->op2, &str[i], j-i);
		//printf("Copied operand2...\n");
		i=j;
	}
	//printf("i, j: %d, %d\n", i, j);

	/* Hunt for operand 3 */
	if (str[i] == ',' || str[i] == ' ') { i++; }
	for (j=i; j<strlen(str); j++) {
		if (str[j] == ' ' || str[j] == ';' || str[j] == ',' || str[j] == '\0') { break; }
	}
	if (j>i) {
		strncpy(instr->op3, &str[i], j-i);
		//printf("Copied operand3...\n");
		i=j;
	}
	//printf("i, j: %d, %d\n", i, j);

	/* Hunt for comment */
	if (str[i] == ' ') { i++; }
	strncpy(instr->comment, &str[i], strlen(str)-i);
	//printf("Copied comment...\n");

	return retval;
}

/* Adds a label to the label list */
int addalabel(char *name, int value) {
	int i;
	int retval = -1;

	/* First character a letter? */
	if (!isalpha(name[0])) {
		return retval;
	}

	/* Label contains only alphanumeric? */
	for (i=0; i<strlen(name); i++) {
		if (!isalnum(name[i])) {
			return retval;
		}
	}

	/* Label already defined? */
	for (i=0; i<labelindex; i++) {
		if (strcmp(label[i].name, name) == 0) {
			return retval;
		}
	}

	if (labelindex == MAXLABELS) {
        printf("ERROR: too many labels! Skipping this label!\n");
        return retval;
	}

	/* Copy new label */
	strncpy(label[labelindex].name, name, LABELLEN);
	label[labelindex].seen = 1;
	label[labelindex].value = value;
	retval=labelindex;
	labelindex++;
	return retval;
}

/* Searches for a label in the pool of labels */
int findalabel(char *name) {
	int i;

	for (i=0; i<labelindex; i++ ) {
		if (strcmp(label[i].name, name) == 0) {
			return i;
		}
	}
	return -1;
}

/* Gest value of label */
int getvaluelabel(int number) {
	return label[number].value;
}

/* Prints all labels */
void printlabels(void) {
	int i;

	if (labelindex>0) {
		printf("Label name       Value\n");
		printf("----------------------\n");
		for (i=0; i<labelindex; i++) {
			printf("%-15s %6d\n", label[i].name, label[i].value);
		}
	} else {
		printf("No lables found.\n");
	}
}

void printmachcode(machinecode_t *mcode) {
	printf("sela: %s ", mcode->sela);
	printf("selb: %s ", mcode->selb);
	printf("seld: %s ", mcode->seld);
	printf("acode: %s ", mcode->acode);
	printf("constant: %s ", mcode->constant);
	printf("end: %c ", mcode->end);
	printf("jcode: %s ", mcode->jcode);
	printf("address: %s\n", mcode->address);
	//printf("%s %s %s %s %s %c %s %s\n", mcode->sela, mcode->selb, mcode->seld, mcode->acode, mcode->constant, mcode->end, mcode->jcode, mcode->address);
}

void writeoutmachcode(FILE* fp, machinecode_t *mcode) {
	fprintf(fp, "\"%s%s%s%s%s%c%s%s\",\n", mcode->sela, mcode->selb, mcode->seld, mcode->acode, mcode->constant, mcode->end, mcode->jcode, mcode->address);
}

int makebinarynumber(char *str, int number, int bits) {
	int i;

	for (i=0; i<bits; i++) {
		if ((number & (1<<i))>0) {
			str[bits-1-i] = '1';
		} else {
			str[bits-1-i] = '0';
		}
	}
	str[bits]='\0';
	//printf("Number: %d\n", number);
	return number;
}

void makenop(machinecode_t *mcode) {
	strcpy(machinecode.sela, "--");
	strcpy(machinecode.selb, "---");
	strcpy(machinecode.seld, "--");
	strcpy(machinecode.acode, "0000");
	strcpy(machinecode.constant, "--------");
	machinecode.end = '0';
	strcpy(machinecode.jcode, "000");
	strcpy(machinecode.address, "-------");
}

int getinteger(char *str) {
	int i=0;
	int j=0;

	if (str[0] == '#') { i=1; }
	if (isdigit(str[i])) {
		return atoi(&str[i]);
	} else {
		j=findalabel(&str[i]);
		//printf("Label found at %d\n", j);
		if (j>=0) {
			return getvaluelabel(j);
		} else {
			printf("ERROR: label not found.\n");
		}
	}
	return -1;
}

int main(int argc, char *argv[]) {

	FILE *infile, *outfile;
	int infileinargv = 1;
	int outfileinargv = 2;
	int verbose = 0;
	char line[LINELEN];
	//int found;
	int pc;
	//int again;
	int lineno;
	time_t seconds;
	int i;

	/* Get time */
   	seconds = time(NULL);


	/* Check if there are at least two parameters */
	if (argc < 3) {
		printf("This is %s\n", VERSION);
		printf("Usage: sasm [--verbose] infile outfile\n");
		return 0;
	}

	/* Check for --verbose */
	if (strcmp(argv[1], "--verbose")==0) {
		printf("DEBUG: enabling verbose mode.\n");
		verbose=1;
		infileinargv++;
		outfileinargv++;
        if (argc < 4) {
            printf("This is %s\n", VERSION);
            printf("Usage: sasm [--verbose] infile outfile\n");
            return 0;
        }
	}

	/* Stop if infile name is outfile name */
	if (strcmp(argv[infileinargv], argv[outfileinargv]) == 0) {
		printf("infile cannot be outfile!\n");
		return 0;
	}

	/* Try to open infile, stop if error */
	if ((infile=fopen(argv[infileinargv], "r")) == NULL) {
		printf("Cannot open infile!\n");
		return 0;
	}

	/* Try to open outfile, stop if error */
	if ((outfile=fopen(argv[outfileinargv], "w")) == NULL) {
		printf("Cannot open outfile!\n");
		fclose(infile);
		return 0;
	}

	/* infile and outfile are now open */
	if (verbose == 1) { printf("DEBUG: infile and outfile are open\n"); }

	//printf("argv's: %s, %s\n", argv[infileinargv], argv[outfileinargv]);

	/********************************************************************/
	/* clear the labels */
	memset(&label, 0, sizeof label);
	/* printf("Sizeof label: %d\n", sizeof label); */

	labelindex = 0;
	pc = 0;
	lineno = 0;

	fprintf(outfile, "-- Assembled by %s on %s", VERSION, ctime(&seconds));

	if (verbose == 1) { printf("DEBUG: entering pass 1...\n"); }
	while (getaline(infile, line, LINELEN) != EOF) {
		lineno++;
		if (verbose == 1) { printf("DEBUG: Parsing line %d...\n", lineno); }

		trimaline(line);
		//printf(">%s<\n", line);
		parsealine(line, &instruction);
		/*printf("label: %s\n", instruction.labelname);*/

		if (strlen(instruction.labelname)>0) {
			if (strcmp(instruction.mnemonic, "equ")==0) {
				if (findalabel(instruction.labelname) >= 0) {
					printf("ERROR: label is already defined in line %d, skipping!\n", lineno);
				} else if (addalabel(instruction.labelname, getinteger(instruction.op1))<0) {
					printf("ERROR: label name contains illegal characters in line %d, skipping!\n", lineno);
				}
			} else {
				if (findalabel(instruction.labelname) >= 0) {
					printf("ERROR: label is already defined in line %d, skipping!\n", lineno);
				} else if (addalabel(instruction.labelname, pc)<0) {
					printf("ERROR: label name contains illegal characters, skipping!\n");
				}
			}
		}

		/* Empty line or EQU */
		if ((strcmp(instruction.mnemonic, "") != 0) && (strcmp(instruction.mnemonic, "equ")!=0)) {
			pc++;
		}
	}
	if (verbose == 1) { printf("DEBUG: found %d labels\n", labelindex); }

	errors=0;
	if (verbose == 1) { printf("DEBUG: entering pass 2...\n"); }
	fseek(infile, 0, SEEK_SET);
	lineno=0;
	pc=0;
	while (getaline(infile, line, LINELEN) != EOF) {
		lineno++;
		if (verbose == 1) { printf("DEBUG: Parsing line %d...\n", lineno); }

		trimaline(line);
		//printf(">%s<\n", line);
		parsealine(line, &instruction);
		if (verbose == 1) {printf("DEBUG: Label: |%s|, mnemonic: |%s|, op1: |%s|, op2: |%s|, op3: |%s|, comment: |%s|\n", instruction.labelname, instruction.mnemonic, instruction.op1, instruction.op2, instruction.op3, instruction.comment); }

		/* Set machinecode to NOP */
		makenop(&machinecode);

		/* Empty line */
		if (strcmp(instruction.mnemonic, "") == 0) {
			continue;
		/* EQUate */
		} else if (strcmp(instruction.mnemonic, "equ") == 0) {
			continue;
		/* NOP */
		} else if (strcmp(instruction.mnemonic, "nop") == 0) {
			//printf("DEBUG: nop found.\n");
		/* LD */
		} else if (strcmp(instruction.mnemonic, "ld") == 0) {
			//printf("DEBUG: ld found.\n");
			strcpy(machinecode.acode, "0001");
			machinecode.end = '1';
			if (strcmp(instruction.op1, "R0")==0 || strcmp(instruction.op1, "r0")==0) {
				strcpy(machinecode.seld, "00");
			} else if (strcmp(instruction.op1,"R1")==0 || strcmp(instruction.op1,"r1")==0) {
				strcpy(machinecode.seld, "01");
			} else if (strcmp(instruction.op1,"R2")==0 || strcmp(instruction.op1,"r2")==0) {
				strcpy(machinecode.seld, "10");
			} else if (strcmp(instruction.op1,"R3")==0 || strcmp(instruction.op1,"r3")==0) {
				strcpy(machinecode.seld, "11");
			} else if (strcmp(instruction.op1,"F")==0 || strcmp(instruction.op1,"f")==0) {
				machinecode.end = '0';
				strcpy(machinecode.seld, "--");
				strcpy(machinecode.acode, "1100");
			} else {
				printf("ERROR: illegal destination in line %d\n", lineno);
				makenop(&machinecode);
				errors++;
			}

			if (strcmp(instruction.op2, "R0")==0 || strcmp(instruction.op2, "r0")==0) {
				strcpy(machinecode.selb, "000");
			} else if (strcmp(instruction.op2,"R1")==0 || strcmp(instruction.op2,"r1")==0) {
				strcpy(machinecode.selb, "001");
			} else if (strcmp(instruction.op2,"R2")==0 || strcmp(instruction.op2,"r2")==0) {
				strcpy(machinecode.selb, "010");
			} else if (strcmp(instruction.op2,"R3")==0 || strcmp(instruction.op2,"r3")==0) {
				strcpy(machinecode.selb, "011");
			} else if (strcmp(instruction.op2,"IN")==0 || strcmp(instruction.op2,"in")==0 || strcmp(instruction.op2,"In")==0 || strcmp(instruction.op2,"iN")==0) {
				strcpy(machinecode.selb, "100");
			} else if (strcmp(instruction.op2,"F")==0 || strcmp(instruction.op2,"f")==0) {
				//machinecode.end = '0';
				//strcpy(machinecode.seld, "--");
				strcpy(machinecode.acode, "1101");
			} else if (instruction.op2[0]=='#') {
				strcpy(machinecode.selb, "101");
				if (makebinarynumber(machinecode.constant, getinteger(instruction.op2), 8)>255) {
					printf("WARNING: number truncated to 8 bits in line %d.\n", lineno);
				}
			} else {
				printf("ERROR: illegal source in line  %d\n", lineno);
				makenop(&machinecode);
				errors++;
			}
		} else if ((strcmp(instruction.mnemonic, "and")==0) || (strcmp(instruction.mnemonic, "or")==0) || (strcmp(instruction.mnemonic, "exor")==0) || (strcmp(instruction.mnemonic, "add")==0) || (strcmp(instruction.mnemonic, "sub")==0)) {
			//printf("DEBUG: %s found.\n", instruction.mnemonic);
			if (strcmp(instruction.mnemonic, "and")==0) {
				strcpy(machinecode.acode, "0100");
			} else if (strcmp(instruction.mnemonic, "or")==0) {
				strcpy(machinecode.acode, "0101");
			} else if (strcmp(instruction.mnemonic, "exor")==0) {
				strcpy(machinecode.acode, "0110");
			} else if (strcmp(instruction.mnemonic, "add")==0) {
				strcpy(machinecode.acode, "0010");
			} else if (strcmp(instruction.mnemonic, "sub")==0) {
				strcpy(machinecode.acode, "0011");
			}
			machinecode.end = '1';
			if (strcmp(instruction.op1, "R0")==0 || strcmp(instruction.op1, "r0")==0) {
				strcpy(machinecode.seld, "00");
			} else if (strcmp(instruction.op1,"R1")==0 || strcmp(instruction.op1,"r1")==0) {
				strcpy(machinecode.seld, "01");
			} else if (strcmp(instruction.op1,"R2")==0 || strcmp(instruction.op1,"r2")==0) {
				strcpy(machinecode.seld, "10");
			} else if (strcmp(instruction.op1,"R3")==0 || strcmp(instruction.op1,"r3")==0) {
				strcpy(machinecode.seld, "11");
			} else {
				printf("ERROR: illegal destination in line %d\n", lineno);
				makenop(&machinecode);
				errors++;
			}
			if (strcmp(instruction.op2, "R0")==0 || strcmp(instruction.op2, "r0")==0) {
				strcpy(machinecode.sela, "00");
			} else if (strcmp(instruction.op2,"R1")==0 || strcmp(instruction.op2,"r1")==0) {
				strcpy(machinecode.sela, "01");
			} else if (strcmp(instruction.op2,"R2")==0 || strcmp(instruction.op2,"r2")==0) {
				strcpy(machinecode.sela, "10");
			} else if (strcmp(instruction.op2,"R3")==0 || strcmp(instruction.op2,"r3")==0) {
				strcpy(machinecode.sela, "11");
			} else {
				printf("ERROR: illegal source 1 in line %d\n", lineno);
				makenop(&machinecode);
				errors++;
			}
			if (strcmp(instruction.op3, "R0")==0 || strcmp(instruction.op3, "r0")==0) {
				strcpy(machinecode.selb, "000");
			} else if (strcmp(instruction.op3,"R1")==0 || strcmp(instruction.op3,"r1")==0) {
				strcpy(machinecode.selb, "001");
			} else if (strcmp(instruction.op3,"R2")==0 || strcmp(instruction.op3,"r2")==0) {
				strcpy(machinecode.selb, "010");
			} else if (strcmp(instruction.op3,"R3")==0 || strcmp(instruction.op3,"r3")==0) {
				strcpy(machinecode.selb, "011");
			} else if (strcmp(instruction.op3,"IN")==0 || strcmp(instruction.op3,"in")==0 || strcmp(instruction.op3,"In")==0 || strcmp(instruction.op3,"iN")==0) {
				strcpy(machinecode.selb, "100");
			} else if (instruction.op3[0]=='#') {
				strcpy(machinecode.selb, "101");
				if (makebinarynumber(machinecode.constant, getinteger(instruction.op3), 8)>255) {
					printf("WARNING: number truncated to 8 bits in line %d\n", lineno);
				}
			} else {
				printf("ERROR: illegal source 2 in line %d\n", lineno);
				makenop(&machinecode);
				errors++;
			}
		} else if (strcmp(instruction.mnemonic, "cmp")==0) {
			strcpy(machinecode.acode, "0011");
			machinecode.end = '0';
			if (strcmp(instruction.op1, "R0")==0 || strcmp(instruction.op1, "r0")==0) {
				strcpy(machinecode.sela, "00");
			} else if (strcmp(instruction.op1,"R1")==0 || strcmp(instruction.op1,"r1")==0) {
				strcpy(machinecode.sela, "01");
			} else if (strcmp(instruction.op1,"R2")==0 || strcmp(instruction.op1,"r2")==0) {
				strcpy(machinecode.sela, "10");
			} else if (strcmp(instruction.op1,"R3")==0 || strcmp(instruction.op1,"r3")==0) {
				strcpy(machinecode.sela, "11");
			} else {
				printf("ERROR: illegal destination in line %d\n", lineno);
				makenop(&machinecode);
				errors++;
			}
			if (strcmp(instruction.op2, "R0")==0 || strcmp(instruction.op2, "r0")==0) {
				strcpy(machinecode.selb, "000");
			} else if (strcmp(instruction.op2,"R1")==0 || strcmp(instruction.op2,"r1")==0) {
				strcpy(machinecode.selb, "001");
			} else if (strcmp(instruction.op2,"R2")==0 || strcmp(instruction.op2,"r2")==0) {
				strcpy(machinecode.selb, "010");
			} else if (strcmp(instruction.op2,"R3")==0 || strcmp(instruction.op2,"r3")==0) {
				strcpy(machinecode.selb, "011");
			} else if (strcmp(instruction.op2,"IN")==0 || strcmp(instruction.op2,"in")==0 || strcmp(instruction.op2,"In")==0 || strcmp(instruction.op2,"iN")==0) {
				strcpy(machinecode.selb, "100");
			} else if (instruction.op2[0]=='#') {
				strcpy(machinecode.selb, "101");
				if (makebinarynumber(machinecode.constant, getinteger(instruction.op2), 8)>255) {
					printf("WARNING: number truncated to 8 bits in line %d.\n", lineno);
				}
			} else {
				printf("ERROR: illegal source1 %s in line %d\n", instruction.op2, lineno);
				makenop(&machinecode);
				errors++;
			}
		} else if ((strcmp(instruction.mnemonic, "shl")==0) || (strcmp(instruction.mnemonic, "shr")==0) || (strcmp(instruction.mnemonic, "rol")==0) || (strcmp(instruction.mnemonic, "ror")==0) || (strcmp(instruction.mnemonic, "not")==0)) {
			//printf("DEBUG: %s found.\n", instruction.mnemonic);
			machinecode.end = '1';
			if (strcmp(instruction.mnemonic, "shl")==0) {
				strcpy(machinecode.acode, "1000");
			} else if (strcmp(instruction.mnemonic, "shr")==0) {
				strcpy(machinecode.acode, "1001");
			} else if (strcmp(instruction.mnemonic, "rol")==0) {
				strcpy(machinecode.acode, "1010");
			} else if (strcmp(instruction.mnemonic, "ror")==0) {
				strcpy(machinecode.acode, "1011");
			} else if (strcmp(instruction.mnemonic, "not")==0) {
				strcpy(machinecode.acode, "0111");
			}
			if (strcmp(instruction.op1, "R0")==0 || strcmp(instruction.op1, "r0")==0) {
				strcpy(machinecode.seld, "00");
			} else if (strcmp(instruction.op1,"R1")==0 || strcmp(instruction.op1,"r1")==0) {
				strcpy(machinecode.seld, "01");
			} else if (strcmp(instruction.op1,"R2")==0 || strcmp(instruction.op1,"r2")==0) {
				strcpy(machinecode.seld, "10");
			} else if (strcmp(instruction.op1,"R3")==0 || strcmp(instruction.op1,"r3")==0) {
				strcpy(machinecode.seld, "11");
			} else {
				printf("ERROR: illegal destination in line %d\n", lineno);
				makenop(&machinecode);
				errors++;
			}
			if (strcmp(instruction.op2, "R0")==0 || strcmp(instruction.op2, "r0")==0) {
				strcpy(machinecode.selb, "000");
			} else if (strcmp(instruction.op2,"R1")==0 || strcmp(instruction.op2,"r1")==0) {
				strcpy(machinecode.selb, "001");
			} else if (strcmp(instruction.op2,"R2")==0 || strcmp(instruction.op2,"r2")==0) {
				strcpy(machinecode.selb, "010");
			} else if (strcmp(instruction.op2,"R3")==0 || strcmp(instruction.op2,"r3")==0) {
				strcpy(machinecode.selb, "011");
			} else if (strcmp(instruction.op2,"IN")==0 || strcmp(instruction.op2,"in")==0 || strcmp(instruction.op2,"In")==0 || strcmp(instruction.op2,"iN")==0) {
				strcpy(machinecode.selb, "100");
			} else if (instruction.op2[0]=='#') {
				strcpy(machinecode.selb, "101");
				if (makebinarynumber(machinecode.constant, getinteger(instruction.op2), 8)>255) {
					printf("WARNING: number truncated to 8 bits in line %d.\n", lineno);
				}
			} else {
				printf("ERROR: illegal source1 %s in line %d\n", instruction.op2, lineno);
				makenop(&machinecode);
				errors++;
			}
		} else if ((strcmp(instruction.mnemonic, "jn")==0) || (strcmp(instruction.mnemonic, "jmp")==0) || (strcmp(instruction.mnemonic, "je")==0) || (strcmp(instruction.mnemonic, "jne")==0) || (strcmp(instruction.mnemonic, "jl")==0) || (strcmp(instruction.mnemonic, "jge")==0) || (strcmp(instruction.mnemonic, "jle")==0) || (strcmp(instruction.mnemonic, "jg")==0) || (strcmp(instruction.mnemonic, "jnz")==0) || (strcmp(instruction.mnemonic, "jz")==0) || (strcmp(instruction.mnemonic, "jnc")==0) || (strcmp(instruction.mnemonic, "jc")==0) || (strcmp(instruction.mnemonic, "jncz")==0)  || (strcmp(instruction.mnemonic, "jcz")==0) || (strcmp(instruction.mnemonic, "j")==0)) {
			//printf("DEBUG: %s found.\n", instruction.mnemonic);
			machinecode.end = '0';
			if (strcmp(instruction.mnemonic, "jn")==0) {
				strcpy(machinecode.jcode, "000");
			} else if (strcmp(instruction.mnemonic, "jne")==0 || (strcmp(instruction.mnemonic, "jnz")==0)) {
				strcpy(machinecode.jcode, "001");
			} else if (strcmp(instruction.mnemonic, "je")==0  || (strcmp(instruction.mnemonic, "jz")==0)) {
				strcpy(machinecode.jcode, "010");
			} else if (strcmp(instruction.mnemonic, "jge")==0 || (strcmp(instruction.mnemonic, "jnc")==0)) {
				strcpy(machinecode.jcode, "011");
			} else if (strcmp(instruction.mnemonic, "jl")==0 || (strcmp(instruction.mnemonic, "jc")==0)) {
				strcpy(machinecode.jcode, "100");
			} else if (strcmp(instruction.mnemonic, "jg")==0 || (strcmp(instruction.mnemonic, "jncz")==0)) {
				strcpy(machinecode.jcode, "101");
			} else if (strcmp(instruction.mnemonic, "jle")==0 || (strcmp(instruction.mnemonic, "jcz")==0)) {
				strcpy(machinecode.jcode, "110");
			} else if (strcmp(instruction.mnemonic, "jmp")==0 || (strcmp(instruction.mnemonic, "j")==0)) {
				strcpy(machinecode.jcode, "111");
			}
			if (instruction.op1[0]=='@') {
				if (makebinarynumber(machinecode.address, getinteger(instruction.op1), 7)>127) {
					printf("WARNING: number truncated to 7 bits in line %d.\n", lineno);
				}
			} else {
				if ((i=findalabel(instruction.op1))>=0) {
					//printf("Label found\n");
					makebinarynumber(machinecode.address, getvaluelabel(i), 7);
				} else {
					printf("ERROR: Unknown jump label in line %d.\n", lineno);
					makenop(&machinecode);
					errors++;
				}
			}
		} else {
			printf("ERROR: Illegal mnemonic in line %d.\n", lineno);
			makenop(&machinecode);
			errors++;
		}
		if (verbose == 1) { printf("DEBUG: "); printmachcode(&machinecode); }
		writeoutmachcode(outfile, &machinecode);
		pc++;
	}

	printlabels();
	printf("Written %d instructions to file %s.\n", pc, argv[outfileinargv]);
	if (errors>0) {
        printf("Errors: %d.\n", errors);
	} else {
        printf("No errors.\n");
    }

	fclose(infile);
	fclose(outfile);
	return 0;
}
